--------------------------------------------------------------------------------
-- Part 2: The Simply-Typed Lambda Calculus with ⊤                            --
--------------------------------------------------------------------------------

module Exam.Part2.Preservation where

-- Imports ---------------------------------------------------------------------

open import Data.Sum using (_⊎_; inj₁; inj₂)
open import Relation.Binary.PropositionalEquality using (_≡_; refl)

open import Exam.Part2.Specification
open import Exam.Util.Fin using (Fin; zero; suc)

-- Operators -------------------------------------------------------------------

infix  4  _⦂_⇒ᵣ_  _⦂_⇒ₛ_

-- Preservation ----------------------------------------------------------------

-- Renamings

-- A renaming ρ has "type" Γ₁ ⇒ Γ₂ if it replaces variables, which in Γ₁
-- have some type t, with variables, which in Γ₂ have the same type t.
_⦂_⇒ᵣ_ : ∀ {m n} → Ren m n → Context m → Context n → Set
ρ ⦂ Γ₁ ⇒ᵣ Γ₂ = ∀ x → Γ₂ ⊢ ` (ρ x) ⦂ lookup Γ₁ x

-- Extending a renaming preserves the typing of the renaming.
⊢extᵣ : ∀ {m n t ρ} {Γ₁ : Context m} {Γ₂ : Context n}
  → ρ      ⦂ Γ₁       ⇒ᵣ Γ₂ 
  → extᵣ ρ ⦂ (Γ₁ , t) ⇒ᵣ (Γ₂ , t)
⊢extᵣ = {!!}

-- Renaming preserves typing, i.e. applying a well-typed renaming to a
-- well-typed term, yields a well-typed term.
⊢ren : ∀ {m n} {Γ₁ : Context m} {Γ₂ : Context n} {e t ρ}
  → ρ ⦂ Γ₁ ⇒ᵣ Γ₂ 
  → Γ₁ ⊢ e ⦂ t
  → Γ₂ ⊢ ren ρ e ⦂ t
⊢ren = {!!}

-- The weakening rename is well-typed.
⊢wk : ∀ {n} {Γ : Context n} {t : Type}
  → wk ⦂ Γ ⇒ᵣ (Γ , t)
⊢wk = {!!}

-- Substitutions

-- A substitution σ has "type" Γ₁ ⇒ Γ₂ if it replaces variables, which in Γ₁
-- have some type t, with terms, which in Γ₂ have the same type t.
_⦂_⇒ₛ_ : ∀ {m n} → Sub m n → Context m → Context n → Set
σ ⦂ Γ₁ ⇒ₛ Γ₂ = ∀ x → Γ₂ ⊢ σ x ⦂ lookup Γ₁ x

-- Extending a substitution preserves the typing of the substitution.
⊢extₛ : ∀ {m n t σ} {Γ₁ : Context m} {Γ₂ : Context n}
  → σ      ⦂ Γ₁       ⇒ₛ Γ₂ 
  → extₛ σ ⦂ (Γ₁ , t) ⇒ₛ (Γ₂ , t)
⊢extₛ = {!!}

-- Substitution preserves typing, i.e. applying a well-typed substitution
-- to a well-typed term, yields a well-typed term.
⊢sub : ∀ {m n} {Γ₁ : Context m} {Γ₂ : Context n} {e t σ}
  → σ ⦂ Γ₁ ⇒ₛ Γ₂ 
  → Γ₁ ⊢ e ⦂ t
  → Γ₂ ⊢ sub σ e ⦂ t
⊢sub = {!!}

-- If a term is well-typed, then the singleton-substitution for that term is well-typed.
⊢0↦ : ∀ {n} {Γ : Context n} {e : Term n} {t}
  → Γ ⊢ e ⦂ t
  → 0↦ e ⦂ (Γ , t) ⇒ₛ Γ
⊢0↦ = {!!}

-- Preservation

preservation : ∀ {e e' t}
  → ∅ ⊢ e ⦂ t
  → e –→ e'
  → ∅ ⊢ e' ⦂ t
preservation = {!!}

