{-# OPTIONS --allow-unsolved-metas #-}

module Exam.Part3.Confluence where

open import Data.Product renaming (_,_ to ⟨_,_⟩)
open import Data.Sum
open import Data.Nat using (ℕ; zero; suc; _+_)
open import Function using (_∘_)
open import Relation.Binary.PropositionalEquality using (_≡_; refl; sym; subst; cong; cong₂; module ≡-Reasoning)
open import Relation.Nullary using (¬_)
open import Data.Empty using (⊥; ⊥-elim)

open import Exam.Part3.Specification
open import Exam.Part3.Composition using (swap-0↦-extᵣ-fusion; swap-0↦-extₛ-fusion)
open import Exam.Util.Fin using (Fin; zero; suc)

--------------------------------------------------------------------------------

-- Congruency rules, but using ↪* instead of ↪

ξ*-λ : ∀ {n} {e e' : Term (suc n)}
  → e ↪* e'
  ---------------
  → `λ e ↪* `λ e'
ξ*-λ = {!!}

ξ*-·₁ : ∀ {n} {e₁ e₂ e₁' : Term n}
  → e₁ ↪* e₁'
  ---------------------
  → e₁ · e₂ ↪* e₁' · e₂ 
ξ*-·₁ = {!!}

ξ*-·₂ : ∀ {n} {e₁ e₂ e₂' : Term n}
  → e₂ ↪* e₂'
  ---------------------
  → e₁ · e₂ ↪* e₁ · e₂'
ξ*-·₂ = {!!}

ξ*-∀₁ : ∀ {n} {t₁ t₁' : Term n} {t₂ : Term (suc n)}
  → t₁ ↪* t₁'
  -------------------------------
  → ∀[x⦂ t₁ ] t₂ ↪* ∀[x⦂ t₁' ] t₂
ξ*-∀₁ = {!!}

ξ*-∀₂ : ∀ {n} {t₁ : Term n} {t₂ t₂' : Term (suc n)}
  → t₂ ↪* t₂'
  -------------------------------
  → ∀[x⦂ t₁ ] t₂ ↪* ∀[x⦂ t₁ ] t₂'
ξ*-∀₂ = {!!}

--------------------------------------------------------------------------------

-- Parallel Reduction

infix   3  _↪ₚ_
data _↪ₚ_ : ∀ {n} → Term n → Term n → Set where
  ξ-` : ∀ {n} {x : Fin n} →
    ` x ↪ₚ ` x
  β-λ : ∀ {n} {e₁ e₁' : Term (suc n)} {e₂ e₂' : Term n} →
    e₁ ↪ₚ e₁' →
    e₂ ↪ₚ e₂' →
    (`λ e₁) · e₂ ↪ₚ e₁' [ e₂' ]
  ξ-λ : ∀ {n} {e e' : Term (suc n)} →
    e ↪ₚ e' →
    `λ e ↪ₚ `λ e'
  ξ-∀ : ∀ {n} {t₁ t₁' : Term n} {t₂ t₂' : Term (suc n)} →
    t₁ ↪ₚ t₁' →
    t₂ ↪ₚ t₂' →
    ∀[x⦂ t₁ ] t₂ ↪ₚ ∀[x⦂ t₁' ] t₂'
  ξ-· : ∀ {n} {e₁ e₁' e₂ e₂' : Term n} →
    e₁ ↪ₚ e₁' →
    e₂ ↪ₚ e₂' →
    e₁ · e₂ ↪ₚ e₁' · e₂'
  ξ-Set : ∀ {n} →
    `Set ↪ₚ (`Set {n = n})

--------------------------------------------------------------------------------

-- Reflexive, transitive closure of parallel reduction

infix   3  _↪ₚ*_
data _↪ₚ*_ : ∀ {n} → Term n → Term n → Set where
  ↪ₚ*-refl : ∀ {n} {e : Term n}
    → e ↪ₚ* e
  ↪ₚ*-step : ∀ {n} {e₁ e₂ e₃ : Term n}
    → e₁ ↪ₚ  e₂
    → e₂ ↪ₚ* e₃
    → e₁ ↪ₚ* e₃

↪ₚ*-trans : ∀ {n} {e₁ e₂ e₃ : Term n}
  → e₁ ↪ₚ* e₂
  → e₂ ↪ₚ* e₃
  → e₁ ↪ₚ* e₃
↪ₚ*-trans = {!!}

module ↪ₚ*-Reasoning where
  infix 1 begin_
  infixr 2 _↪ₚ⟨_⟩_ _↪ₚ*⟨_⟩_ _≡⟨_⟩_ _≡⟨⟩_
  infix 3 _∎

  begin_ : ∀ {n} {e₁ e₂ : Term n} → e₁ ↪ₚ* e₂ → e₁ ↪ₚ* e₂
  begin p = p

  _↪ₚ⟨_⟩_ : ∀ {n} (e₁ {e₂} {e₃} : Term n) → e₁ ↪ₚ e₂ → e₂ ↪ₚ* e₃ → e₁ ↪ₚ* e₃
  _ ↪ₚ⟨ p ⟩ q = ↪ₚ*-step p q

  _↪ₚ*⟨_⟩_ : ∀ {n} (e₁ {e₂} {e₃} : Term n) → e₁ ↪ₚ* e₂ → e₂ ↪ₚ* e₃ → e₁ ↪ₚ* e₃
  _ ↪ₚ*⟨ p ⟩ q = ↪ₚ*-trans p q

  _≡⟨_⟩_ : ∀ {n} (e₁ {e₂} {e₃} : Term n) → e₁ ≡ e₂ → e₂ ↪ₚ* e₃ → e₁ ↪ₚ* e₃
  _ ≡⟨ refl ⟩ q = q

  _≡⟨⟩_ : ∀ {n} (e₁ {e₂} {e₃} : Term n) → e₁ ↪ₚ* e₂ → e₁ ↪ₚ* e₂
  _ ≡⟨⟩ q = q

  _∎ : ∀ {n} (e : Term n) → e ↪ₚ* e
  _ ∎ = ↪ₚ*-refl

--------------------------------------------------------------------------------

-- Parallel reduction is reflexive

↪ₚ-refl : ∀ {n} {t : Term n}
  → t ↪ₚ t
↪ₚ-refl = {!!}

--------------------------------------------------------------------------------

-- Regular reduction implies parallel reduction

↪→↪ₚ : ∀ {n} {t t' : Term n}
  → t ↪ t'
  → t ↪ₚ t'
↪→↪ₚ = {!!}

↪*→↪ₚ* : ∀ {n} {t t' : Term n}
  → t ↪* t'
  → t ↪ₚ* t'
↪*→↪ₚ* = {!!}

--------------------------------------------------------------------------------

-- Parallel reduction implies regular reduction

↪ₚ→↪* : ∀ {n} {t t' : Term n}
  → t ↪ₚ t'
  → t ↪* t'
↪ₚ→↪* = {!!}

↪ₚ*→↪* : ∀ {n} {t t' : Term n}
  → t ↪ₚ* t'
  → t ↪* t'
↪ₚ*→↪* = {!!}

--------------------------------------------------------------------------------

-- Parallel reduction is preserved under renaming

↪ₚ-ren :
  ∀ {m n} {t t' : Term m} (ρ : Ren m n) →
  t ↪ₚ t' →
  ren ρ t ↪ₚ ren ρ t'
↪ₚ-ren = {!!}

--------------------------------------------------------------------------------

-- A substitution σ₁ reduces to σ₂, if each term of σ₁ reduces to the
-- corresponding term of σ₂.
_↪ₚσ_ : ∀ {m n} (σ₁ σ₂ : Sub m n) → Set
σ₁ ↪ₚσ σ₂ = ∀ x → σ₁ x ↪ₚ σ₂ x

↪ₚσ-refl : ∀ {m n} {σ : Sub m n} →
  σ ↪ₚσ σ
↪ₚσ-refl = {!!}

↪ₚσ-ext :  ∀ {m n} {σ σ' : Sub m n}
  → σ ↪ₚσ σ'
  → extₛ σ ↪ₚσ extₛ σ'
↪ₚσ-ext = {!!}

↪ₚσ-sub : ∀ {m n} {e e' : Term m} {σ σ' : Sub m n} →
  σ ↪ₚσ σ' →
  e ↪ₚ e' →
  sub σ e ↪ₚ sub σ' e'
↪ₚσ-sub = {!!}

↪ₚσ-0↦ : ∀ {n} {e₁ e₂ : Term n} →
  e₁ ↪ₚ e₂ →
  (0↦ e₁) ↪ₚσ (0↦ e₂)
↪ₚσ-0↦ = {!!}

↪ₚσ-[] : ∀ {n} {e₁ e₁' : Term (suc n)}  {e₂ e₂' : Term n} →
  e₁ ↪ₚ e₁' →
  e₂ ↪ₚ e₂' →
  e₁ [ e₂ ] ↪ₚ e₁' [ e₂' ]
↪ₚσ-[] = {!!}

--------------------------------------------------------------------------------

_⁺ : ∀ {n} → Term n → Term n
(` x) ⁺               = ` x
(`λ e) ⁺              = `λ (e ⁺)
((`λ e₁) · e₂) ⁺      = e₁ ⁺ [ e₂ ⁺ ]
(e₁ · e₂) ⁺           = e₁ ⁺ · (e₂ ⁺)
(∀[x⦂ t₁ ] t₂) ⁺      = ∀[x⦂ (t₁ ⁺) ] (t₂ ⁺)
`Set ⁺                = `Set

-- Note, in part 4, this proof might have *many* cases. That's to be
-- expected, and most of them are actually very similar.

par-triangle : ∀ {n} {e e' : Term n}
  → e ↪ₚ e'
    -------
  → e' ↪ₚ (e ⁺)
par-triangle = {!!}

strip : ∀ {n} {e e₁ e₂ : Term n}
  → e ↪ₚ e₁
  → e ↪ₚ* e₂
  → ∃[ e' ] (e₁ ↪ₚ* e') × (e₂ ↪ₚ e')
strip = {!!}

-- Confluence for parallel reduction.

confluenceₚ : ∀ {n} {e e₁ e₂ : Term n}
  → e ↪ₚ* e₁
  → e ↪ₚ* e₂
  → ∃[ e' ] (e₁ ↪ₚ* e') × (e₂ ↪ₚ* e')
confluenceₚ = {!!}

-- Confluence for our regular reduction. This is what we actually want.

confluence : ∀ {n} {e e₁ e₂ : Term n}
  → e ↪* e₁
  → e ↪* e₂
  → ∃[ e' ]
      e₁ ↪* e' ×
      e₂ ↪* e'
confluence = {!!}
